import matplotlib.pyplot as plt
import numpy as np

from .annotations import Annotation
from . import scoring


def plotSampleScoring(ref: Annotation, hyp: Annotation, fs: int = 1) -> plt.figure:
    """Build an overview plot showing the outcome of sample scoring.

    Args:
        ref (Annotation): Reference annotations (ground - truth)
        hyp (Annotation): Hypotheses annotations (output of a ML pipeline)
        fs (int): Sampling frequency of the labels. Default 1 Hz.

    Returns:
        plt.figure: Output matplotlib figure
    """

    score = scoring.SampleScoring(ref, hyp, fs)
    time = np.arange(len(score.tpMask)) / fs
    # Resample Data
    ref = Annotation(ref.events, fs, round(len(ref.mask) / ref.fs * fs))
    hyp = Annotation(hyp.events, fs, round(len(hyp.mask) / hyp.fs * fs))

    fig = plt.figure(figsize=(16, 3))

    # Plot background shading
    plt.fill_between(time, 0, 1, where=score.tpMask,
                     alpha=0.2, color='tab:green',
                     transform=plt.gca().get_xaxis_transform())
    plt.fill_between(time, 0, 1, where=score.fnMask,
                     alpha=0.2, color='tab:purple',
                     transform=plt.gca().get_xaxis_transform())
    plt.fill_between(time, 0, 1, where=score.fpMask,
                     alpha=0.2, color='tab:red',
                     transform=plt.gca().get_xaxis_transform())

    # Plot Labels
    plt.plot(time, ref.mask * 0.4 + 0.6, 'k')
    plt.plot(time, hyp.mask * 0.4 + 0.1, 'k')

    # Plot Colored dots for detections
    lineFn, = plt.plot(time[score.fnMask], score.fnMask[score.fnMask], 'o', color='tab:purple')
    lineTp, = plt.plot(time[score.tpMask], score.tpMask[score.tpMask], 'o', color='tab:green')
    plt.plot(time[score.tpMask], score.tpMask[score.tpMask] * 0.5, 'o', color='tab:green')
    lineFp, = plt.plot(time[score.fpMask], score.fpMask[score.fpMask] * 0.5, 'o', color='tab:red')

    # Text
    plt.title('Sample based Scoring')

    plt.yticks([0.3, 0.8], ['HYP', 'REF'])
    _scale_time_xaxis(fig)

    _buildLegend(lineTp, lineFn, lineFp, score, fig)

    return fig


def plotEventScoring(ref: Annotation, hyp: Annotation,
                     param: scoring.EventScoring.Parameters = scoring.EventScoring.Parameters()) -> plt.figure:
    """Build an overview plot showing the outcome of event scoring.

    Args:
        ref (Annotation): Reference annotations (ground - truth)
        hyp (Annotation): Hypotheses annotations (output of a ML pipeline)
        param(EventScoring.Parameters, optional):  Parameters for event scoring.
            Defaults to default values.

    Returns:
        plt.figure: Output matplotlib figure
    """
    def _plotEvent(x, y, color):
        plt.axvspan(x[0], x[1], alpha=0.2, color=color)
        if x[1] - x[0] > 0:
            plt.plot(x, y, color=color, linewidth=5, solid_capstyle='butt')
        else:
            plt.scatter(x[0], y[0], color=color)

    score = scoring.EventScoring(ref, hyp, param)
    time = np.arange(len(ref.mask)) / ref.fs

    fig = plt.figure(figsize=(16, 3))

    # Plot Labels
    plt.plot(time, ref.mask * 0.4 + 0.6, 'k')
    plt.plot(time, hyp.mask * 0.4 + 0.1, 'k')

    # Initialize lines for legend
    lineTp, = plt.plot([], [], color='tab:green', linewidth=5)
    lineFn, = plt.plot([], [], color='tab:purple', linewidth=5)
    lineFp, = plt.plot([], [], color='tab:red', linewidth=5)

    # Plot REF TP & FN
    for event in ref.events:
        # TP
        if np.any(score.tpMask[round(event[0] * score.fs):round(event[1] * score.fs)]):
            _plotEvent([event[0], event[1] - (1 / ref.fs)], [1, 1], 'tab:green')
            score.tpMask[round(event[0] * score.fs):round(event[1] * score.fs)] = 1
        else:
            _plotEvent([event[0], event[1] - (1 / ref.fs)], [1, 1], 'tab:purple')

    # Plot HYP TP & FP
    for event in hyp.events:
        # FP
        if np.all(~score.tpMask[round(event[0] * score.fs):round(event[1] * score.fs)]):
            _plotEvent([event[0], event[1] - (1 / ref.fs)], [0.5, 0.5], 'tab:red')
        # TP
        elif np.all(score.tpMask[round(event[0] * score.fs):round(event[1] * score.fs)]):
            plt.plot([event[0], event[1] - (1 / ref.fs)], [0.5, 0.5],
                     color='tab:green', linewidth=5, solid_capstyle='butt', linestyle='solid')
        # Mix TP, FP
        else:
            _plotEvent([event[0], event[1] - (1 / ref.fs)], [0.5, 0.5], 'tab:red')
            plt.plot([event[0], event[1] - (1 / ref.fs)], [0.5, 0.5],
                     color='tab:green', linewidth=5, solid_capstyle='butt', linestyle=(0, (2, 2)))

    # Text
    plt.title('Event Scoring')

    plt.yticks([0.3, 0.8], ['HYP', 'REF'])
    _scale_time_xaxis(fig)

    _buildLegend(lineTp, lineFn, lineFp, score, fig)

    return fig


def _scale_time_xaxis(fig: plt.figure):
    """Scale x axis of a figure where initial values are in seconds.

    The function leaves the xaxis as is if the number of seconds to display is < 5 * 60
    If it is larger than 5 minutes, xaxis is formatted as m:s
    If it is larger than 5 hours, xaxis is formatted as h:m:s

    Args:
        fig (plt.figure): figure to handle
    """

    def s2m(x, _):
        return f'{int(x / 60)}:{int(x%60)}'

    def s2h(x, _):
        return f'{int(x / 3600)}:{int((x / 60)%60)}:{int(x%60)}'

    maxTime = fig.gca().get_xlim()[1]
    if maxTime > 5 * 60 * 60:
        fig.gca().xaxis.set_major_formatter(s2h)
        fig.gca().set_xlabel('time [h:m:s]')
    elif maxTime > 5 * 60:
        fig.gca().xaxis.set_major_formatter(s2m)
        fig.gca().set_xlabel('time [m:s]')
    else:
        fig.gca().set_xlabel('time [s]')


def _buildLegend(lineTp, lineFn, lineFp, score, fig):
    """Build legend and adjust spacing for scoring text"""
    plt.legend([lineTp, lineFn, lineFp],
               ['TP: {}'.format(np.sum(score.tp)),
                'FN: {}'.format(np.sum(score.refTrue - score.tp)),
                'FP: {}'.format(np.sum(score.fp))], loc=(1.02, 0.65))

    textstr = "• Sensitivity: {:.2f}\n".format(score.sensitivity)
    textstr += "• Precision  : {:.2f}\n".format(score.precision)
    textstr += "• F1 - score   : {:.2f}".format(score.f1)
    fig.text(1.02, 0.05, textstr, fontsize=12, transform=plt.gca().transAxes)

    # Adjust spacing
    plt.margins(x=0)  # No margin on X data
    plt.tight_layout()
    fig.subplots_adjust(right=0.86)  # Allow space for scoring text


if __name__ == "__main__":
    fs = 1
    duration = 66 * 60
    ref = Annotation([(8 * 60, 12 * 60), (30 * 60, 35 * 60), (48 * 60, 50 * 60)], fs, duration)
    hyp = Annotation([(8 * 60, 12 * 60), (28 * 60, 32 * 60), (50.5 * 60, 51 * 60), (60 * 60, 62 * 60)], fs, duration)

    fig = plotSampleScoring(ref, hyp)
    param = scoring.EventScoring.Parameters(
        toleranceStart=30,
        toleranceEnd=60,
        minOverlap=0,
        maxEventDuration=5 * 60,
        minDurationBetweenEvents=90)
    fig = plotEventScoring(ref, hyp, param)
    plt.show()



''' Unit testing
'''

__author__ = "Jonathan Dan"
__email__ = "jonathan.dan at epfl.ch"

import unittest

import numpy as np

from src.timescoring.annotations import Annotation
from src.timescoring import scoring


class TestAnnotation(unittest.TestCase):
    def assertListOfTupleEqual(expected, actual, message):
        difference = set(expected) ^ set(actual)
        assert not difference, 'List of events check failed : {}'.format(message)

    def assertMask(expected, actual, message):
        difference = np.sum(expected ^ actual)
        assert not difference, 'Mask check failed : {}'.format(message)

    def checkMaskEvents(mask, events, fs, numSamples, message):
        labels = Annotation(events, fs, numSamples)
        TestAnnotation.assertMask(mask, labels.mask, message)

        labels = Annotation(mask, fs)
        TestAnnotation.assertListOfTupleEqual(events, labels.events, message)

    def test_mask_to_eventList(self):
        fs = 10
        numSamples = 10

        # Simple events
        mask = [0, 1, 1, 0, 0, 0, 1, 1, 1, 0]
        events = [(0.1, 0.3), (0.6, 0.9)]
        TestAnnotation.checkMaskEvents(mask, events, fs, numSamples, 'Simple events')

        # Event == File duration
        mask = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
        events = [(0.0, 1.0)]
        TestAnnotation.checkMaskEvents(mask, events, fs, numSamples, 'Event = file duration')

        # Event at start
        mask = [1, 1, 1, 1, 0, 0, 1, 0, 0, 0]
        events = [(0.0, 0.4), (0.6, 0.7)]
        TestAnnotation.checkMaskEvents(mask, events, fs, numSamples, 'event at start')

        # Event at end
        mask = [0, 1, 1, 0, 0, 0, 0, 1, 1, 1]
        events = [(0.1, 0.3), (0.7, 1.)]
        TestAnnotation.checkMaskEvents(mask, events, fs, numSamples, 'event at end')

        # Event at start and end
        mask = [1, 1, 1, 0, 0, 0, 0, 1, 0, 1]
        events = [(0.0, 0.3), (0.7, 0.8), (0.9, 1.0)]
        TestAnnotation.checkMaskEvents(mask, events, fs, numSamples, 'event at start and end')


class TestSampleScoring(unittest.TestCase):
    def test_sample_scoring(self):
        fs = 1

        # Simple events
        ref = Annotation([1, 1, 1, 0, 0, 0, 1, 1, 1, 0], fs)
        hyp = Annotation([0, 1, 1, 0, 1, 1, 0, 0, 1, 0], fs)
        scores = scoring.SampleScoring(ref, hyp)
        np.testing.assert_equal(scores.sensitivity, 3 / 6, 'sensitivity simple test')
        np.testing.assert_equal(scores.precision, 3 / 5, 'precision simple test')
        np.testing.assert_equal(scores.fpRate, 2 * 3600 * 24 / 10, 'FP / day simple test')

        # No detections
        ref = Annotation([1, 1, 1, 0, 0, 0, 1, 1, 1, 0], fs)
        hyp = Annotation([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], fs)
        scores = scoring.SampleScoring(ref, hyp)
        np.testing.assert_equal(scores.sensitivity, 0, 'sensitivity no detections')
        np.testing.assert_equal(scores.precision, np.nan, 'precision no detections')
        np.testing.assert_equal(scores.fpRate, 0, 'FP / day no detections')

        # No events
        ref = Annotation([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], fs)
        hyp = Annotation([0, 1, 1, 0, 1, 1, 0, 0, 1, 0], fs)
        scores = scoring.SampleScoring(ref, hyp)
        np.testing.assert_equal(scores.sensitivity, np.nan, 'sensitivity no events')
        np.testing.assert_equal(scores.precision, 0, 'precision no events')
        np.testing.assert_equal(scores.fpRate, 5 * 3600 * 24 / 10, 'FP / day no events')

        # Resampling
        fs = 256
        ref = Annotation([(1.2, 5.4), (7.1, 8.7)], fs, fs * 10)
        hyp = Annotation([(3, 5), (6.8, 9.2)], fs, fs * 10)
        scores = scoring.SampleScoring(ref, hyp)
        np.testing.assert_equal(scores.sensitivity, 4 / 6, 'sensitivity resampling')
        np.testing.assert_equal(scores.precision, 1, 'precision resampling')
        np.testing.assert_equal(scores.fpRate, 0, 'FP / day resampling')


class TestEventScoring(unittest.TestCase):
    def test_long_event_splitting(self):
        fs = 10
        numSamples = 10
        events = [(0.0, 0.5), (0.6, 0.9)]
        maxEventDuration = 0.2

        labels = Annotation(events, fs, numSamples)
        shortLabels = scoring.EventScoring._splitLongEvents(labels, maxEventDuration)

        # Check maximum duration
        for event in shortLabels.events:
            self.assertLessEqual(event[1] - event[0], maxEventDuration + 1e-10,  # Computre precision tolerance
                                 "Long event splitting resulted in long event")

        # Check mask remains unchanged
        TestAnnotation.assertMask(labels.mask, shortLabels.mask,
                                  "Event mask changed when splitting events")

    def test_merge_neighbouring_events(self):
        fs = 10
        labels = Annotation([(1, 2), (2, 3), (3.5, 4), (4, 5), (18, 20)], fs, 600)
        minDurationBetweenEvents = 0.2

        mergedEvents = scoring.EventScoring._mergeNeighbouringEvents(labels, minDurationBetweenEvents)
        expectedMergedEvents = Annotation([(1, 3), (3.5, 5), (18, 20)], fs, 600)
        TestAnnotation.assertListOfTupleEqual(expectedMergedEvents.events, mergedEvents.events, 'Test merge events')

    def test_extending_events(self):
        fs = 1
        numSamples = 10
        events = [(0.0, 2.0), (6.0, 8.0)]
        before = 0.5
        after = 1.0

        target = [(0.0, 3.0), (5.5, 9.0)]

        labels = Annotation(events, fs, numSamples)
        extendedLabels = scoring.EventScoring._extendEvents(labels, before, after)

        TestAnnotation.assertListOfTupleEqual(target, extendedLabels.events,
                                              "Extended events mismatch")

    def test_event_scoring(self):
        fs = 10
        numSamples = 60 * 60 * fs  # 1 hour

        # Simple events
        ref = Annotation([(40, 60)], fs, numSamples)
        hyp = Annotation([(10, 20), (42, 65)], fs, numSamples)
        param = scoring.EventScoring.Parameters(toleranceStart=0,
                                                toleranceEnd=10,
                                                minDurationBetweenEvents=0)
        scores = scoring.EventScoring(ref, hyp, param)
        np.testing.assert_equal(scores.sensitivity, 1, 'sensitivity no detections')
        np.testing.assert_equal(scores.precision, 0.5, 'precision no detections')
        np.testing.assert_equal(scores.fpRate, 1 * 24, 'FP / day no detections')

        # Tolerance before events
        # REF      <----->
        # HYP    <------->
        ref = Annotation([(40, 60)], fs, numSamples)
        hyp = Annotation([(39, 60)], fs, numSamples)
        param = scoring.EventScoring.Parameters(toleranceStart=1)
        scores = scoring.EventScoring(ref, hyp, param)
        np.testing.assert_equal(scores.sensitivity, 1, 'sensitivity no detections')
        np.testing.assert_equal(scores.precision, 1, 'precision no detections')
        np.testing.assert_equal(scores.fpRate, 0, 'FP / day no detections')

        # Split long events
        # REF <----->
        # HYP   <-------------------------->
        # SPLIT  <----------------><-------->
        ref = Annotation([(40, 60)], fs, numSamples)
        hyp = Annotation([(42, 65 + 6 * 60)], fs, numSamples)
        param = scoring.EventScoring.Parameters(maxEventDuration=5 * 60)
        scores = scoring.EventScoring(ref, hyp, param)
        np.testing.assert_equal(scores.sensitivity, 1, 'sensitivity no detections')
        np.testing.assert_equal(scores.precision, 1 / 3, 'precision no detections')
        np.testing.assert_equal(scores.fpRate, 2 * 24, 'FP / day no detections')

        # No detections
        ref = Annotation([(40, 60)], fs, numSamples)
        hyp = Annotation([], fs, numSamples)
        scores = scoring.EventScoring(ref, hyp)
        np.testing.assert_equal(scores.sensitivity, 0, 'sensitivity no detections')
        np.testing.assert_equal(scores.precision, np.nan, 'precision no detections')
        np.testing.assert_equal(scores.fpRate, 0, 'FP / day no detections')

        # No events
        ref = Annotation([], fs, numSamples)
        hyp = Annotation([(40, 60)], fs, numSamples)
        scores = scoring.EventScoring(ref, hyp)
        np.testing.assert_equal(scores.sensitivity, np.nan, 'sensitivity no events')
        np.testing.assert_equal(scores.precision, 0, 'precision no events')
        np.testing.assert_equal(scores.fpRate, 1 * 24, 'FP / day no events')

        #
        # Seizure with typical distribution, some overlapping some not
        ref = Annotation([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                          0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0,
                          0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                         fs)
        hyp = Annotation([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
                          1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0,
                          0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1],
                         fs)
        param = scoring.EventScoring.Parameters(
            toleranceStart=0,
            toleranceEnd=0,
            minOverlap=0,
            maxEventDuration=5 * 60,
            minDurationBetweenEvents=0)
        scores = scoring.EventScoring(ref, hyp, param)
        np.testing.assert_equal(scores.sensitivity, 2 / 3, 'sensitivity typical distribution')
        np.testing.assert_equal(scores.precision, 0.4, 'precision typical distribution')

        param.minOverlap = 0.5
        scores = scoring.EventScoring(ref, hyp, param)
        np.testing.assert_equal(scores.sensitivity, 1 / 3, 'sensitivity typical distribution')
        np.testing.assert_equal(scores.precision, 1 / 4, 'precision typical distribution')

        #
        # Long hypothesis seizure spaning several true seizures
        message = 'Long predicted seizure spaning several true seizures'
        ref = Annotation([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0,
                          0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0,
                          0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                         fs)
        hyp = Annotation([0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0,
                          0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0,
                          0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                         fs)

        param.minOverlap = 0.0
        scores = scoring.EventScoring(ref, hyp, param)
        np.testing.assert_equal(scores.sensitivity, 1, 'sensitivity : ' + message)
        np.testing.assert_equal(scores.precision, 5 / 7, 'precision : ' + message)

        #
        # Long true seizure and many short predicted seizures​
        message = 'Long true seizure and many short predicted seizures​'
        ref = Annotation([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                          1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0,
                          0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                         fs)
        hyp = Annotation([0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0,
                          0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                          0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0],
                         fs)

        param.minOverlap = 0.0
        scores = scoring.EventScoring(ref, hyp, param)
        np.testing.assert_equal(scores.sensitivity, 0.5, 'sensitivity : ' + message)
        np.testing.assert_equal(scores.precision, 0.25, 'precision : ' + message)

        # Typial distribution (one missed extended REF would hide FP)
        fs = 1
        ref = Annotation([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                          0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0,
                          0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                         fs)
        hyp = Annotation([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
                          1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0,
                          0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1],
                         fs)
        message = 'typical - extended REF would hide FP'
        param.minOverlap = 0.6
        scores = scoring.EventScoring(ref, hyp, param)
        np.testing.assert_equal(scores.sensitivity, 1 / 3, 'sensitivity : ' + message)
        np.testing.assert_equal(scores.precision, 0.25, 'precision : ' + message)


if __name__ == '__main__':
    unittest.main()



''' Class to store annotations both as a binary mask and as a list of events.
'''

__author__ = "Jonathan Dan"
__email__ = "jonathan.dan at epfl.ch"

from dataclasses import dataclass
from typing import List, Tuple

import numpy as np
from nptyping import Bool, NDArray, Shape


@dataclass(frozen=True)
class Annotation:
    """Class to store annotations both as a binary mask and as a list of events.

    Instances can be initiated either from a a binary mask or from a list of events,
    the other data representation will be automatically generated. The class properties
    are immutable to guarantee consistency between both data types.
    """
    events: List[Tuple[int, int]]
    mask: NDArray[Shape["Size"], Bool]
    fs: int

    def __init__(self, data, fs: int, numSamples: int = None):
        """Initialize an annotation instance.
        - Annotation(mask, fs):
        This can either be done by providing a binary vector where positive labels are
        indicated by True.

        - Annotation(events, fs, numSamples)
        Or by provding a list of (start, stop) tuples for each event. Start and stop
        times are expected in seconds.

        The annotation class contains two immutable fields: mask and events.

        Args:
            data (List[Tuple[int, int]] OR NDArray[Bool]): _description_
            fs (int): Sampling frequency in Hertz of the annotations.
            numSamples (int, optional): Is required when initalizing by providing a
                list of (start, stop) tuples. It indicates the number of annotation
                samples in the annotation binary mask. It should be left to None if
                a binary mask is provided. Defaults to None.

        Raises:
            TypeError: Raises a TypeError if input does not meet one of the expected
                data types.
        """
        object.__setattr__(self, 'fs', fs)  # Write to frozen object
        # Annotation(events, fs, numSamples)
        # Init by providing a list of start, stop tuples for each event
        if numSamples is not None:
            # Build binary mask associated with list of events
            mask = np.zeros((numSamples, ), dtype=np.bool_)
            for event in data:
                mask[round(event[0] * fs):round(event[1] * fs)] = True
            object.__setattr__(self, 'events', data)  # Write to frozen object
            object.__setattr__(self, 'mask', mask)  # Write to frozen object

        # Annotation(mask, fs)
        # Init provided by a binary mask with True labels during event
        elif numSamples is None:
            events = list()
            tmpEnd = []
            # Find transitions
            start_i = np.where(np.diff(np.array(data, dtype=int)) == 1)[0]
            end_i = np.where(np.diff(np.array(data, dtype=int)) == -1)[0]

            # No transitions and first sample is positive -> event is duration of file
            if len(start_i) == 0 and data[0]:
                events.append((0, len(data) / fs))
            else:
                # Edge effect - First sample is an event
                if data[0]:
                    events.append((0, (end_i[0] + 1) / fs))
                    end_i = np.delete(end_i, 0)
                # Edge effect - Last event runs until end of file
                if data[-1]:
                    if len(start_i):
                        tmpEnd = [((start_i[-1] + 1) / fs, len(data) / fs)]
                        start_i = np.delete(start_i, len(start_i) - 1)
                # Add all events
                start_i += 1
                end_i += 1
                for i in range(len(start_i)):
                    events.append((start_i[i] / fs, end_i[i] / fs))
                events += tmpEnd  # add potential end edge effect
            object.__setattr__(self, 'events', events)  # Write to frozen object
            object.__setattr__(self, 'mask', np.array(data, dtype=np.bool_))  # Write to frozen object

        else:
            raise TypeError
